import csv
import json
import glob
import os
import math
from girder_worker.core.io import fetch
import networkx as nx
from collections import namedtuple
from networkx.algorithms.shortest_paths.generic import all_shortest_paths
from networkx.algorithms.shortest_paths.unweighted import (
    single_source_shortest_path
)


conv_graph = nx.DiGraph()


class Validator(namedtuple('Validator', ['type', 'format'])):
    """Validator

    .. py:attribute:: type

        The validator type, like ``string``.

    .. py:attribute:: format

        The validator format, like ``text``.
    """

    def is_valid(self):
        """Return whether the type/format combination is valid.

        If format is None, checks for the presence of any valid type/format with
        the specified type.

        :returns: ``True`` if ``type`` and ``format`` are a valid, loaded
            type/format pair.
        """
        if self.format is None:
            return self.type in set(v.type for v in conv_graph.nodes())

        return self in conv_graph.nodes()


def get_csv_reader(input):

    # csv package does not support unicode
    input = str(input)

    # Special case: detect single-column files.
    # This check assumes that our only valid delimiters are commas and tabs.
    firstLine = input.split('\n')[0]
    if not ('\t' in firstLine or ',' in firstLine) \
            or len(input.splitlines()) == 1:
        dialect = 'excel'

    else:
        # Take a data sample to determine dialect, but
        # don't include incomplete last line
        sample = ''
        sampleSize = 0
        while len(sample) == 0:
            sampleSize += 5000
            sample = '\n'.join(input[:sampleSize].splitlines()[:-1])
        dialect = csv.Sniffer().sniff(sample)
        dialect.skipinitialspace = True
    return csv.DictReader(input.splitlines(), dialect=dialect)


def csv_to_rows(input):
    reader = get_csv_reader(input)
    rows = [d for d in reader]
    fields = reader.fieldnames

    output = {'fields': fields, 'rows': rows}

    # Attempt numeric conversion
    for row in output['rows']:
        for col in row:
            try:
                row[col] = int(row[col])
            except Exception:
                try:
                    orig = row[col]
                    row[col] = float(row[col])

                    # Disallow NaN, Inf, -Inf since this does not
                    # pass through JSON converters cleanly
                    if math.isnan(row[col]) or math.isinf(row[col]):
                        row[col] = orig
                except Exception:
                    pass

    return output


def converter_path(source, target):
    """Gives the shortest path that should be taken to go from a source
    type/format to a target type/format.

    Throws a ``NetworkXNoPath`` exception if it can not find a path.

    :param source: Validator tuple indicating the type/format being converted
        `from`.
    :param target: ``Validator`` tuple indicating the type/format being
        converted `to`.
    :returns: An ordered list of the analyses that need to be run to convert
        from ``source`` to ``target``.
    """
    # These are to ensure an exception gets thrown if source/target don't exist
    get_validator_analysis(source)
    get_validator_analysis(target)

    # We sort and pick the first of the shortest paths just to produce a stable
    # conversion path. This is stable in regards to which plugins are loaded at
    # the time.
    paths = all_shortest_paths(conv_graph, source, target)
    path = sorted(paths)[0]
    path = zip(path[:-1], path[1:])

    return [conv_graph.edge[u][v] for (u, v) in path]


def has_converter(source, target=Validator(type=None, format=None)):
    """Determines if any converters exist from a given type, and optionally format.

    Underneath, this just traverses the edges until it finds one which matches
    the arguments.

    :param source: ``Validator`` tuple indicating the type/format being
        converted `from`.
    :param target: ``Validator`` tuple indicating the type/format being
        converted `to`.
    :returns: ``True`` if it can converter from ``source`` to ``target``,
        ``False`` otherwise.
    """
    sources = []

    for node in conv_graph.nodes():
        if ((source.type is None) or (source.type == node.type)) and \
           ((source.format is None) or (source.format == node.format)):
            sources.append(node)

    for u in sources:
        reachable = single_source_shortest_path(conv_graph, u)
        del reachable[u]  # Ignore path to self, since there are no self loops

        for v in reachable:
            if ((target.type is None) or (target.type == v.type)) and \
               ((target.format is None) or (target.format == v.format)):
                return True

    return False


def get_validator_analysis(validator):
    """Gets a validator's analysis from the conversion graph.

    >>> analysis = get_validator_analysis(Validator('string', 'text'))

    Returns an analysis dictionary

    >>> type(analysis) == dict
    True

    Which contains an inputs key

    >>> 'inputs' in analysis
    True

    If the validator doesn't exist, an exception will be raised

    >>> get_validator_analysis(Validator('foo', 'bar'))
    Traceback (most recent call last):
    ...
    Exception: No such validator foo/bar

    :param validator: A ``Validator`` namedtuple
    :returns: A dictionary containing the runnable analysis.
    """
    try:
        return conv_graph.node[validator]
    except KeyError:
        raise Exception(
            'No such validator %s/%s' % (validator.type, validator.format))


def import_converters(search_paths):
    """
    Import converters and validators from the specified search paths.
    These functions are loaded into ``girder_worker.format.conv_graph`` with
    nodes representing validators, and directed edges representing
    converters.

    Any files in a search path matching ``validate_*.json`` are loaded
    as validators. Validators should be fast (ideally O(1)) algorithms
    for determining if data is of the specified format. These are algorithms
    that have a single input named ``"input"`` and a single output named
    ``"output"``. The input has the type and format to be checked.
    The output must have type and format ``"boolean"``. The script performs
    the validation and sets the output variable to either true or false.

    Any ``*_to_*.json`` files are imported as converters.
    A converter is simply an analysis with one input named ``"input"`` and one
    output named ``"output"``. The input and output should have matching
    type but should be of different formats.

    :param search_paths: A list of search paths relative to the current
        working directory. Passing a single path as a string also works.
    :type search_paths: str or list of str
    """

    if not isinstance(search_paths, (list, tuple)):
        search_paths = [search_paths]

    def get_analysis(filename):
        with open(filename) as f:
            analysis = json.load(f)

            if 'script' not in analysis:
                analysis['script'] = fetch({
                    'mode': analysis.get('script_fetch_mode', 'auto'),
                    'url': analysis['script_uri']
                })

        return analysis

    prevdir = os.getcwd()
    for path in search_paths:
        os.chdir(path)
        validator_files = set(glob.glob(os.path.join(path, 'validate_*.json')))
        converter_files = set(glob.glob(os.path.join(
            path, '*_to_*.json'))) - validator_files

        for filename in validator_files:
            analysis = get_analysis(filename)

            # Validators only contain 1 input and output, so the type/format of
            # it can be gleaned from the first input.
            conv_graph.add_node(Validator(analysis['inputs'][0]['type'],
                                          analysis['inputs'][0]['format']),
                                analysis)

        for filename in converter_files:
            analysis = get_analysis(filename)
            in_type = analysis['inputs'][0]['type']
            in_format = analysis['inputs'][0]['format']
            out_format = analysis['outputs'][0]['format']

            conv_graph.add_edge(Validator(in_type, in_format),
                                Validator(in_type, out_format),
                                attr_dict=analysis)

    os.chdir(prevdir)


def print_conversion_graph():
    """
    Print a graph of supported conversion paths in DOT format to standard
    output.
    """

    print 'digraph g {'

    for node in conv_graph.nodes():
        paths = single_source_shortest_path(conv_graph, node)
        reachable_conversions = [p for p in paths.keys() if p != node]

        for dest in reachable_conversions:
            print '"%s:%s" -> "%s:%s"' % (node[0], node[1], dest[0], dest[1])

    print '}'


def print_conversion_table():
    """
    Print a table of supported conversion paths in CSV format with ``'from'``
    and ``'to'`` columns to standard output.
    """

    print 'from,to'

    for node in conv_graph.nodes():
        paths = single_source_shortest_path(conv_graph, node)
        reachable_conversions = [p for p in paths.keys() if p != node]

        for dest in reachable_conversions:
            print '%s:%s,%s:%s' % (node[0], node[1], dest[0], dest[1])


def import_default_converters():
    """
    Import converters from the default search paths. This is called when the
    :py:mod:`girder_worker.format` module is first loaded.
    """

    cur_path = os.path.dirname(os.path.realpath(__file__))
    import_converters([os.path.join(cur_path, t) for t in [
        'boolean',
        'directory',
        'graph',
        'image',
        'integer',
        'integer_list',
        'netcdf',
        'number',
        'number_list',
        'python',
        'string',
        'string_list',
        'table',
        'tree']])


import_default_converters()



import os
import shutil
import subprocess
import tempfile
import time
from girder_worker import config, logger
from girder_worker.docker.utils import chmod_writable

# Minimum interval in seconds at which to run the docker-gc script
MIN_GC_INTERVAL = 600


def before_run(e):
    import executor
    if e.info['task'].get('mode') == 'docker':
        executor.validate_task_outputs(e.info['task_outputs'])
    if not _read_bool_from_config('gc', False):
        e.info.setdefault('kwargs', {})['_rm_container'] = True


def _read_from_config(key, default):
    """
    Helper to read Docker specific config values from the worker config files.
    """
    if config.has_option('docker', key):
        return config.get('docker', key)
    else:
        return default


def _read_bool_from_config(key, default):
    """
    Helper to read Docker specific bool config values from the worker config files.
    """
    if config.has_option('docker', key):
        return config.getboolean('docker', key)
    else:
        return default


def docker_gc(e):
    """
    Garbage collect containers that have not been run in the last hour using the
    https://github.com/spotify/docker-gc project's script, which is copied in
    the same directory as this file. After that, deletes all images that are
    no longer used by any containers.
    """
    if not _read_bool_from_config('gc', False):
        return
    stampfile = os.path.join(config.get('girder_worker', 'tmp_root'), '.dockergcstamp')
    if os.path.exists(stampfile) and time.time() - os.path.getmtime(stampfile) < MIN_GC_INTERVAL:
        return
    else:  # touch the file
        with open(stampfile, 'w') as f:
            f.write('')

    logger.info('Garbage collecting docker containers and images.')
    gc_dir = tempfile.mkdtemp()

    try:
        script = os.path.join(os.path.dirname(__file__), 'docker-gc')
        if not os.path.isfile(script):
            raise Exception('Docker GC script %s not found.' % script)
        if not os.access(script, os.X_OK):
            raise Exception('Docker GC script %s is not executable.' % script)

        env = os.environ.copy()
        env['FORCE_CONTAINER_REMOVAL'] = '1'
        env['STATE_DIR'] = gc_dir
        env['PID_DIR'] = gc_dir
        env['GRACE_PERIOD_SECONDS'] = str(_read_from_config('cache_timeout', 3600))

        # Handle excluded images
        excluded = _read_from_config('exclude_images', '').split(',')
        excluded = [img for img in excluded if img.strip()]
        if excluded:
            exclude_file = os.path.join(gc_dir, '.docker-gc-exclude')
            with open(exclude_file, 'w') as fd:
                fd.write('\n'.join(excluded) + '\n')
            env['EXCLUDE_FROM_GC'] = exclude_file

        p = subprocess.Popen(args=(script,), env=env)
        p.wait()  # Wait for garbage collection subprocess to finish

        if p.returncode != 0:
            raise Exception('Docker GC returned code %d.' % p.returncode)
    finally:
        shutil.rmtree(gc_dir)


def task_cleanup(e):
    """
    Since files written by docker containers are owned by root, we can't
    clean them up in the worker process since that typically doesn't run
    as root. So, we run a lightweight container to make the temp dir cleanable.
    """
    if e.info['task']['mode'] == 'docker' and '_tempdir' in e.info['kwargs']:
        tmpdir = e.info['kwargs']['_tempdir']
        chmod_writable(tmpdir)


def load(params):
    from girder_worker.core import events, register_executor
    import executor

    events.bind('run.before', params['name'], before_run)
    events.bind('run.finally', params['name'], task_cleanup)
    events.bind('cleanup', params['name'], docker_gc)
    register_executor('docker', executor.run)



def fetch(spec, **kwargs):
    import pymongo
    import bson
    db = spec['db']
    collection = spec['collection']
    host = spec.get('host', 'localhost')
    data = ''.join([bson.BSON.encode(d) for d in
                    pymongo.MongoClient(host)[db][collection].find()])
    return data


def push(data, spec, **kwargs):
    import pymongo
    import bson
    db = spec['db']
    collection = spec['collection']
    host = spec.get('host', 'localhost')
    bson_data = bson.decode_all(data)

    c = pymongo.MongoClient(host)[db][collection]
    # TODO is this really what we want? Dropping the whole collection?
    # Seems dangerous, might be unexpected.
    c.drop()
    if len(bson_data) > 0:
        c.insert(bson_data)



###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder_jobs.constants import JobStatus


class CustomJobStatus(object):
    """
    The custom job status flags for the worker.
    """
    FETCHING_INPUT = 820
    CONVERTING_INPUT = 821
    CONVERTING_OUTPUT = 822
    PUSHING_OUTPUT = 823
    CANCELING = 824

    # valid transitions for worker scheduled jobs
    valid_worker_transitions = {
        JobStatus.QUEUED: [JobStatus.INACTIVE],
        JobStatus.RUNNING: [JobStatus.QUEUED, FETCHING_INPUT],
        FETCHING_INPUT: [JobStatus.RUNNING],
        CONVERTING_INPUT: [JobStatus.RUNNING, FETCHING_INPUT],
        CONVERTING_OUTPUT: [JobStatus.RUNNING],
        PUSHING_OUTPUT: [JobStatus.RUNNING, CONVERTING_OUTPUT],
        CANCELING: [JobStatus.INACTIVE, JobStatus.QUEUED, JobStatus.RUNNING],
        JobStatus.ERROR: [FETCHING_INPUT, CONVERTING_INPUT, CONVERTING_OUTPUT,
                          PUSHING_OUTPUT, CANCELING, JobStatus.QUEUED,
                          JobStatus.RUNNING],
        # The last two are allowed for revoke called from outside Girder
        JobStatus.CANCELED: [CANCELING, JobStatus.QUEUED, JobStatus.RUNNING],
        JobStatus.SUCCESS: [JobStatus.RUNNING, PUSHING_OUTPUT]
    }

    # valid transitions for celery scheduled jobs
    # N.B. We have the extra worker input/output states defined here for when
    # we are running girder_worker.run as a regular celery task
    valid_celery_transitions = {
        JobStatus.QUEUED: [JobStatus.INACTIVE],
        # Note celery tasks can jump straight from INACTIVE to RUNNING
        JobStatus.RUNNING: [JobStatus.INACTIVE, JobStatus.QUEUED,
                            FETCHING_INPUT],
        FETCHING_INPUT: [JobStatus.RUNNING],
        CONVERTING_INPUT: [JobStatus.RUNNING, FETCHING_INPUT],
        CONVERTING_OUTPUT: [JobStatus.RUNNING],
        PUSHING_OUTPUT: [JobStatus.RUNNING, CONVERTING_OUTPUT],
        CANCELING: [JobStatus.INACTIVE, JobStatus.QUEUED, JobStatus.RUNNING],
        JobStatus.ERROR: [FETCHING_INPUT, CONVERTING_INPUT, CONVERTING_OUTPUT,
                          PUSHING_OUTPUT, CANCELING, JobStatus.QUEUED,
                          JobStatus.RUNNING],
        JobStatus.CANCELED: [CANCELING, JobStatus.INACTIVE, JobStatus.QUEUED,
                             JobStatus.RUNNING],
        JobStatus.SUCCESS: [JobStatus.RUNNING, PUSHING_OUTPUT]
    }

    @classmethod
    def isValid(cls, status):
        return status in (
            cls.FETCHING_INPUT,
            cls.CONVERTING_INPUT,
            cls.CONVERTING_OUTPUT,
            cls.PUSHING_OUTPUT,
            cls.CANCELING
        )

    @classmethod
    def validTransitionsWorker(cls, status):
        return cls.valid_worker_transitions.get(status)

    @classmethod
    def validTransitionsCelery(cls, status):
        return cls.valid_celery_transitions.get(status)



import struct
import json


class StreamPushAdapter(object):
    """
    This represents the interface that must be implemented by push adapters for
    IO modes that want to implement streaming output.
    """

    def write(self, buf):
        """
        Write a chunk of data to the output stream.
        """
        raise NotImplemented

    def close(self):
        """
        Close the output stream. Called after the last data is sent.
        """
        pass


class JobProgressAdapter(StreamPushAdapter):
    """
    This reads structured JSON documents one line at a time and sends
    them as progress events via the JobManager.

    :param job_manager: The job manager to use to send the progress events.
    :type job_manager: girder_worker.utils.JobManager
    """

    def __init__(self, job_manager):
        super(JobProgressAdapter, self).__init__()

        self.job_manager = job_manager
        self._buf = b''

    def write(self, buf):
        lines = buf.split(b'\n')
        if self._buf:
            lines[0] = self._buf + lines[0]
        self._buf = lines[-1]

        for line in lines[:-1]:
            self._parse(line)

    def _parse(self, line):
        try:
            doc = json.loads(line.decode('utf8'))
        except ValueError:
            return  # TODO log?

        if not isinstance(doc, dict):
            return  # TODO log?

        self.job_manager.updateProgress(
            total=doc.get('total'), current=doc.get('current'), message=doc.get('message'))


class DockerStreamPushAdapter(StreamPushAdapter):
    """
    An adapter that reads a docker stream. The format is a Header and a Payload (frame)
    where the header has the following format:

    .. code-block:: none

        header := [8]byte{STREAM_TYPE, 0, 0, 0, SIZE1, SIZE2, SIZE3, SIZE4}

    We want to read the header to get the size of the payload, read the payload
    and forward it on to another adapter.
    """
    def __init__(self, adapter):
        self._adapter = adapter
        self._reset()

    def _reset(self):
        self._header = b''
        self._header_bytes_read = 0
        self._payload_bytes_read = 0
        self._payload_size = None

    def _read_header(self):
        """
        Read the header or part of the header. When the head has been read, the
        payload size is decoded and returned, otherwise return None.
        """
        bytes_to_read = min(8 - self._header_bytes_read, self._data_length-self._data_offset)
        self._header += self._data[self._data_offset:self._data_offset+bytes_to_read]
        self._data_offset += bytes_to_read
        self._header_bytes_read += bytes_to_read

        if self._header_bytes_read == 8:
            _, payload_size = struct.unpack('>BxxxL', self._header)

            return payload_size

    def _read_payload(self):
        """
        Read the payload or part of the payload. The data is written directly to
        the wrapped adapter.
        """
        bytes_to_read = min(self._payload_size - self._payload_bytes_read,
                            self._data_length-self._data_offset)
        self._adapter.write(self._data[self._data_offset:self._data_offset+bytes_to_read])
        self._data_offset += bytes_to_read
        self._payload_bytes_read += bytes_to_read

    def write(self, data):
        self._data = data
        self._data_length = len(data)
        self._data_offset = 0

        # While we still have data iterate over it
        while self._data_length > self._data_offset:
            # We are reading the header
            if self._header_bytes_read < 8:
                self._payload_size = self._read_header()

            # We are reading the payload
            if self._payload_size and self._payload_bytes_read < self._payload_size:
                self._read_payload()

            # We are done with this payload
            if self._payload_size == self._payload_bytes_read:
                self._reset()

    def close(self):
        self._adapter.close()



# -*- coding: utf-8 -*-

from Bio.Phylo import BaseTree


def recursive_attr(obj):
    """Recursively parase through phyloXML data that is not a Clade"""
    dictionary = {}
    try:
        # if obj has subobjects, parse throuh them and insert in a dictionary
        for key in obj.__dict__.keys():
            attr = getattr(obj, key)
            if attr and attr is not None:
                if isinstance(attr, BaseTree.TreeElement):
                    dictionary[key] = recursive_attr(attr)
                else:
                    dictionary[key] = getattr(obj, key)
        return dictionary
    except Exception:
        # else just return the object
        return obj


def recursive_clade(obj, data_coll, tree_coll=None):
    """Recursively parse through phyloXML"""
    # instantiate variables
    clade_children = []
    tempDict = {}
    if tree_coll is not None:
        treeDict = {}

    # iterate through each attribute of the obj
    for key in obj.__dict__.keys():
        attr = getattr(obj, key)
        if attr and attr is not None:
            # debugging
            print 'key = %s, attr = %s' % (key, attr)

            # get all non-recursive clade attributes first
            # do we need to even do this first? SZ -Aug 16, 2012
            if isinstance(attr, BaseTree.Clade):
                clade_children.append(attr)
            elif isinstance(attr, list):
                tempDictList = []
                # iterate through each element of the list
                for elem in attr:
                    if isinstance(elem, BaseTree.Clade):
                        clade_children.append(elem)
                    # if it's a complex phyloxml TreeElement attribute, parse
                    # it using recurisve_attr
                    else:
                        print('  elem3, attr = %s, type = %s'
                              % (attr, type(elem)))
                        tempDictList.append(recursive_attr(elem))
                tempDict[key] = tempDictList
            else:  # isinstance(attr, BaseTree.TreeElement):
                tempDict[key] = recursive_attr(attr)
    # process all clades that are children of obj
    if clade_children:
        tempDict['clades'] = []
        if tree_coll is not None:
            treeDict['clades'] = []
        for child in clade_children:
            child_id = recursive_clade(child, data_coll, tree_coll)
            tempDict['clades'].append(child_id['dataId'])

            if tree_coll is not None:
                # if we want to insert dataLink data in the treeDict, this is
                # how we'd do it
                # treeDict['children'].append({'treeLink':child_id['treeId'],
                #                              'dataLink':child_id['dataId']})
                treeDict['clades'].append(child_id['treeId'])
    # if args.debug_level >= 1:
    # print 'object name: ', obj.name, tempDict
    # insert into mongodb here

    # add mongodb id
    tempDict['_id'] = insertIntoMongo(tempDict, data_coll)
    if tree_coll is not None:
        treeDict['dataLink'] = tempDict['_id']
        insertIntoMongo(treeDict, tree_coll)
    # if args.debug_level >= 1:
    # print tempDict
    # CRL - added option for return withou treeDict
    if tree_coll is not None:
        return {'dataId': tempDict['_id'], 'treeId': treeDict['_id']}
    else:
        return {'dataId': tempDict['_id']}


def insertIntoMongo(item, collection):
    return collection.insert(item)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder import events
from girder.constants import AccessType
from girder.plugin import getPlugin, GirderPlugin
from girder_jobs.models.job import Job

from .api.worker import Worker
from . import event_handlers


class WorkerPlugin(GirderPlugin):
    DISPLAY_NAME = 'Worker'
    CLIENT_SOURCE_PATH = 'web_client'

    def load(self, info):
        getPlugin('jobs').load(info)

        info['apiRoot'].worker = Worker()

        events.bind('jobs.schedule', 'worker', event_handlers.schedule)
        events.bind('jobs.status.validate', 'worker', event_handlers.validateJobStatus)
        events.bind('jobs.status.validTransitions', 'worker', event_handlers.validTransitions)
        events.bind('jobs.cancel', 'worker', event_handlers.cancel)
        events.bind('model.job.save.after', 'worker', event_handlers.attachJobInfoSpec)
        events.bind('model.job.save', 'worker', event_handlers.attachParentJob)
        Job().exposeFields(AccessType.SITE_ADMIN, {'celeryTaskId', 'celeryQueue'})



import os
import zipfile

output = input + '.zip'

with zipfile.ZipFile(output, 'w') as zf:
    for root, _, files in os.walk(input):
        for file in files:
            abspath = os.path.join(root, file)
            zf.write(os.path.join(root, file),
                     arcname=os.path.relpath(abspath, os.path.dirname(input)))



import os
import re

from girder_worker.core import TaskSpecValidationError, utils
from girder_worker.core.io import make_stream_fetch_adapter, make_stream_push_adapter
from girder_worker.docker.tasks import _docker_run
from girder_worker.docker.transforms import (
    ContainerStdOut,
    ContainerStdErr
)
from girder_worker.docker.transforms.girder import ProgressPipe
from girder_worker.docker.io import (
    NamedPipeReader,
    NamedPipeWriter,
    FDWriteStreamConnector,
    NamedPipe,
    FDReadStreamConnector
)


DATA_VOLUME = '/mnt/girder_worker/data'

_inputRe = re.compile(r'\$input\{([^}]+)\}')
_outputRe = re.compile(r'\$output\{([^}]+)\}')
_flagRe = re.compile(r'\$flag\{([^}]+)\}')


def _transform_path(inputs, taskInputs, inputId, tmpDir):
    """
    If the input specified by inputId is a filepath target, we transform it to
    its absolute path within the Docker container (underneath the data mount).
    """
    for ti in taskInputs.itervalues():
        tiId = ti['id'] if 'id' in ti else ti['name']
        if tiId == inputId:
            if (ti.get('target') == 'filepath' and
                    inputs[inputId]['script_data'].startswith(tmpDir.rstrip(os.sep) + os.sep)):
                rel = os.path.relpath(inputs[inputId]['script_data'], tmpDir)
                return os.path.join(DATA_VOLUME, rel)
            else:
                return inputs[inputId]['script_data']

    raise Exception('No task input found with id = ' + inputId)


def _expand_args(args, inputs, taskInputs, outputs, tmpDir):
    """
    Expands arguments to the container execution if they reference input or output
    data. For example, if an input has id=foo, then a container arg of the form
    ``$input{foo}`` would be expanded to the runtime value of that input. If that
    input is a filepath target, the file path will be transformed into the
    location that it will be available inside the running container.

    Output file paths can be referenced using ``$output{foo}``. That will expand
    to use the "name" field of the corresponding output binding for this task invocation.
    TODO(zach) add example of task output and corresponding output binding. This
    expansion will use the absolute path format with DATA_VOLUME as the base path.
    """
    newArgs = []

    for arg in args:
        skip = False
        for inputId in _inputRe.findall(arg):
            if inputId in inputs:
                transformed = _transform_path(
                    inputs, taskInputs, inputId, tmpDir)
                arg = arg.replace('$input{%s}' % inputId, str(transformed))
            elif inputId == '_tempdir':
                arg = arg.replace('$input{_tempdir}', DATA_VOLUME)
            else:
                raise Exception('Could not expand token: $input{%s}.' % inputId)
        for inputId in _flagRe.findall(arg):
            if inputId in inputs and inputs[inputId]['script_data']:
                val = taskInputs[inputId].get('arg', inputId)
            else:
                val = ''
            arg = arg.replace('$flag{%s}' % inputId, val)
            if not arg:
                skip = True
        for outputId in _outputRe.findall(arg):
            if outputId in outputs and 'name' in outputs[outputId]:
                arg = arg.replace(
                    '$output{%s}' % outputId, os.path.join(DATA_VOLUME, outputs[outputId]['name']))
            else:
                raise Exception('Could not expand token: $output{%s}.' % outputId)
        if not skip:
            newArgs.append(arg)

    return newArgs


def validate_task_outputs(task_outputs):
    """
    This is called prior to fetching inputs to make sure the output specs are
    valid. Outputs in docker mode can result in side effects, so it's best to
    make sure the specs are valid prior to fetching.
    """
    for name, spec in task_outputs.iteritems():
        if spec.get('target') == 'filepath':
            path = spec.get('path', name)
            if path.startswith('/') and not path.startswith(DATA_VOLUME + '/'):
                raise TaskSpecValidationError(
                    'Docker filepath output paths must either start with '
                    '"%s/" or be specified relative to that directory.' %
                    DATA_VOLUME)
        elif name not in ('_stdout', '_stderr'):
            raise TaskSpecValidationError(
                'Docker outputs must be either "_stdout", "_stderr", or '
                'filepath-target outputs.')


def _setup_streams(task_inputs, inputs, task_outputs, outputs, tempdir, job_mgr, progress_pipe):
    """
    Returns a 2 tuple of input and output pipe mappings. The first element is
    a dict mapping input file descriptors to the corresponding stream adapters,
    the second is a dict mapping output file descriptors to the corresponding
    stream adapters. This also handles the special cases of STDIN, STDOUT, and
    STDERR mappings, and in the case of non-streaming standard IO pipes, will
    create default bindings for those as well.
    """
    stream_connectors = []

    def stream_pipe_path(id, spec, bindings):
        """
        Helper to check for a  valid streaming IO specs. If the
        given spec is not a streaming spec, returns None. If it is, returns the
        path.
        """
        if spec.get('stream') and id in bindings and spec.get('target') == 'filepath':
            path = spec.get('path', id)
            if path.startswith('/'):
                raise Exception('Streaming filepaths must be relative.')
            path = os.path.join(tempdir, path)
            return path

        return None

    # handle stream inputs
    for id, spec in task_inputs.iteritems():

        path = stream_pipe_path(id, spec, inputs)
        # We have a streaming input
        if path is not None:
            writer = NamedPipeWriter(NamedPipe(path))
            connector = FDWriteStreamConnector(make_stream_fetch_adapter(inputs[id]), writer)
            stream_connectors.append(connector)
            # Don't open from this side, must be opened for reading first!

    # handle stream outputs
    for id, spec in task_outputs.iteritems():
        path = stream_pipe_path(id, spec, outputs)
        if path is not None:
            reader = NamedPipeReader(NamedPipe(path))
            connector = FDReadStreamConnector(reader, make_stream_push_adapter(outputs[id]))
            stream_connectors.append(connector)

    # handle special stream output for job progress
    if progress_pipe and job_mgr:
        progress_pipe = ProgressPipe(os.path.join(tempdir, '.girder_progress'))
        stream_connectors.append(progress_pipe.open())

    return stream_connectors


def _setup_std_streams(task_outputs, outputs):
    def push_adapter(id):
        adapter = None
        if id in task_outputs and id in outputs:
            if task_outputs[id].get('stream'):
                adapter = make_stream_push_adapter(outputs[id])
            else:
                adapter = utils.AccumulateDictAdapter(outputs[id], 'script_data')

        return adapter

    return (push_adapter('_stdout'), push_adapter('_stderr'))


def add_input_volumes(inputs, volumes):
    """
    For any filepath input that has a direct_path property AND a script_data
    property that matches, add a read-only volume to the docker volumes record
    to give access to that path.

    :param inputs: the spec inputs.
    :param volumes: the docker volumes dictionary to modify.
    """
    for input in inputs.itervalues():
        if 'direct_path' in input and input.get('script_data') == input['direct_path']:
            volume = input.get('script_data')
            if volume not in volumes:
                volumes[volume] = {
                    'bind': volume,
                    'mode': 'ro'
                }


def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    image = task['docker_image']
    celery_task = kwargs.get('_celery_task')
    pull_image = task.get('pull_image', True)
    progress_pipe = task.get('progress_pipe', False)
    remove_container = kwargs.get('_rm_container', False)
    # Allow run args to overridden
    extra_run_kwargs = task.get('docker_run_args', {})
    tempdir = kwargs.get('_tempdir')
    job_mgr = kwargs.get('_job_manager')
    args = _expand_args(task.get('container_args', []), inputs, task_inputs, outputs, tempdir)

    stream_connectors = _setup_streams(
        task_inputs, inputs, task_outputs, outputs, tempdir, job_mgr, progress_pipe)

    stdout_fetch_adapter, stderr_fetch_adapter = _setup_std_streams(task_outputs, outputs)

    # Connect up stdout and strerr ContainerStdOut() and ContainerStdErr() will be
    # replaced with the real streams when the container is started.
    if stdout_fetch_adapter is not None:
        stream_connectors.append(FDReadStreamConnector(ContainerStdOut(), stdout_fetch_adapter))
    if stderr_fetch_adapter is not None:
        stream_connectors.append(FDReadStreamConnector(ContainerStdErr(), stderr_fetch_adapter))

    entrypoint = None
    if 'entrypoint' in task:
        if isinstance(task['entrypoint'], (list, tuple)):
            entrypoint = task['entrypoint']
        else:
            entrypoint = [task['entrypoint']]

    volumes = {
        tempdir: {
            'bind': DATA_VOLUME,
            'mode': 'rw'
        }
    }
    add_input_volumes(inputs, volumes)

    _docker_run(celery_task, image, pull_image=pull_image, entrypoint=entrypoint,
                container_args=args, volumes=volumes, remove_container=remove_container,
                stream_connectors=stream_connectors, **extra_run_kwargs)

    for name, spec in task_outputs.iteritems():
        if spec.get('target') == 'filepath' and not spec.get('stream'):
            path = spec.get('path', '$output{%s}' % name)
            for outputId in _outputRe.findall(path):
                if outputId in outputs and 'name' in outputs[outputId]:
                    path = path.replace('$output{%s}' % outputId, outputs[outputId]['name'])
                elif 'path' in spec:
                    raise Exception('Could not expand token: $output{%s}.' % outputId)
                else:
                    path = name

            if not path.startswith('/'):
                # Assume relative paths are relative to the data volume
                path = os.path.join(DATA_VOLUME, path)

            # Convert data volume refs to the temp dir on the host
            path = path.replace(DATA_VOLUME, tempdir, 1)
            if not os.path.exists(path):
                raise Exception('Output filepath %s does not exist.' % path)
            outputs[name]['script_data'] = path



import os
import tarfile


output = os.path.splitext(input)[0]

try:
    os.makedirs(output)
except OSError:
    if not os.path.exists(output):
        raise

with tarfile.open(input, 'r') as tf:
    tf.extractall(output)



import httplib
import os
import re
import requests
import six
import ssl
import urlparse

from girder_worker.core.utils import StreamFetchAdapter, StreamPushAdapter


class HttpStreamFetchAdapter(StreamFetchAdapter):
    def __init__(self, input_spec):
        super(HttpStreamFetchAdapter, self).__init__(input_spec)

        self._iter = None  # will be lazily created

    def read(self, buf_len):
        """
        Implementation note: due to a constraint of the requests library, the
        buf_len that is used the first time this method is called will cause
        all future requests to ``read`` to have the same ``buf_len`` even if
        a different ``buf_len`` is passed in on subsequent requests.
        """
        if self._iter is None:  # lazy load response body iterator
            method = self.input_spec.get('method', 'GET').upper()
            headers = self.input_spec.get('headers', {})
            params = self.input_spec.get('params', {})
            req = requests.request(
                method, self.input_spec['url'], headers=headers, params=params,
                stream=True, allow_redirects=True)
            req.raise_for_status()  # we have the response headers already
            self._iter = req.iter_content(buf_len, decode_unicode=False)

        try:
            return six.next(self._iter)
        except StopIteration:
            return b''


class HttpStreamPushAdapter(StreamPushAdapter):
    def __init__(self, output_spec):
        """
        Uses HTTP chunked transfer-encoding to stream a request body to a
        server. Unfortunately requests does not support hooking into this logic
        easily, so we use the lower-level httplib module.
        """
        super(HttpStreamPushAdapter, self).__init__(output_spec)
        self._closed = False

        parts = urlparse.urlparse(output_spec['url'])
        if parts.scheme == 'https':
            ssl_context = ssl.create_default_context()
            conn = httplib.HTTPSConnection(parts.netloc, context=ssl_context)
        else:
            conn = httplib.HTTPConnection(parts.netloc)

        try:
            conn.putrequest(output_spec.get('method', 'POST').upper(),
                            parts.path, skip_accept_encoding=True)

            for header, value in output_spec.get('headers', {}).items():
                conn.putheader(header, value)

            conn.putheader('Transfer-Encoding', 'chunked')
            conn.endheaders()  # This actually flushes the headers to the server
        except Exception:
            print('HTTP connection to "%s" failed.' % output_spec['url'])
            conn.close()
            raise

        self.conn = conn

    def write(self, buf):
        """
        Write a chunk of data to the output stream in accordance with the
        chunked transfer encoding protocol.
        """
        try:
            self.conn.send(hex(len(buf))[2:].encode('utf-8'))
            self.conn.send(b'\r\n')
            self.conn.send(buf)
            self.conn.send(b'\r\n')
        except Exception:
            resp = self.conn.getresponse()
            print('Exception while sending HTTP chunk to %s, status was %s, '
                  'message was:\n%s' % (self.output_spec['url'], resp.status,
                                        resp.read()))
            self.conn.close()
            self._closed = True
            raise

    def close(self):
        """
        Close the output stream. Called after the last data is sent.
        """
        if self._closed:
            return

        try:
            self.conn.send(b'0\r\n\r\n')
            resp = self.conn.getresponse()
            if resp.status >= 300 and resp.status < 400:
                raise Exception('Redirects are not supported for streaming '
                                'requests at this time. %d to Location: %s' % (
                                    resp.status, resp.getheader('Location')))
            if resp.status >= 400:
                raise Exception(
                    'HTTP stream output to %s failed with status %d. Response '
                    'was: %s' % (
                        self.output_spec['url'], resp.status, resp.read()))
        finally:
            self.conn.close()


def _read_filename_from_resp(request, url):
    """
    This helper will derive a filename from the HTTP response, first attempting
    to use the content disposition header, otherwise falling back to the last
    token of the URL.
    """
    match = re.search('filename="(.*)"',
                      request.headers.get('Content-Disposition', ''))

    if match is None:
        return [t for t in url.split('/') if t][-1]
    else:
        return match.group(1)


def fetch(spec, **kwargs):
    """
    Downloads an input file via HTTP using requests.
    """
    if 'url' not in spec:
        raise Exception('No URL specified for HTTP input.')
    task_input = kwargs.get('task_input', {})
    target = task_input.get('target', 'memory')
    url = spec['url']
    method = spec.get('method', 'GET').upper()
    request = requests.request(method, url, headers=spec.get('headers', {}),
                               params=spec.get('params', {}),
                               stream=True, allow_redirects=True)

    try:
        request.raise_for_status()
    except Exception:
        print 'HTTP fetch failed (%s). Response: %s' % (url, request.text)
        raise

    if target == 'filepath':
        tmpDir = kwargs['_tempdir']

        if 'filename' in task_input:
            filename = task_input['filename']
        else:
            filename = _read_filename_from_resp(request, url)

        path = os.path.join(tmpDir, filename)

        total = 0
        maxSize = spec.get('maxSize')

        with open(path, 'wb') as out:
            for buf in request.iter_content(65536):
                length = len(buf)
                if maxSize and length + total > maxSize:
                    raise Exception(
                        'Exceeded max download size of %d bytes.' % maxSize)
                out.write(buf)
                total += length

        return path
    elif target == 'memory':
        return ''.join(request.iter_content(65536))
    else:
        raise Exception('Invalid HTTP fetch target: ' + target)


def push(data, spec, **kwargs):
    task_output = kwargs.get('task_output', {})
    target = task_output.get('target', 'memory')

    url = spec['url']
    method = spec.get('method', 'POST').upper()

    if target == 'filepath':
        with open(data, 'rb') as fd:
            request = requests.request(
                method, url, headers=spec.get('headers', {}), data=fd,
                params=spec.get('params', {}), allow_redirects=True)
    elif target == 'memory':
        request = requests.request(
            method, url, headers=spec.get('headers', {}), data=data,
            params=spec.get('params', {}), allow_redirects=True)
    else:
        raise Exception('Invalid HTTP fetch target: ' + target)

    try:
        request.raise_for_status()
    except Exception:
        print 'HTTP push failed (%s). Response: %s' % (url, request.text)
        raise



def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):  # noqa
    from girder_worker.core import events, io, run as core_run, utils

    # Make map of steps
    steps = {step['name']: step for step in task['steps']}

    # Make map of input bindings
    bindings = {step['name']: {} for step in task['steps']}

    # Create dependency graph and downstream pointers
    dependencies = {step['name']: set() for step in task['steps']}
    downstream = {}
    for conn in task['connections']:
        # Add dependency graph link for internal links
        if 'input_step' in conn and 'output_step' in conn:
            dependencies[conn['input_step']].add(conn['output_step'])

        # Add downstream links for links with output
        if 'output_step' in conn:
            ds = downstream.setdefault(conn['output_step'], {})
            ds_list = ds.setdefault(conn['output'], [])
            ds_list.append(conn)

        # Set initial bindings for inputs
        if 'input_step' in conn and 'output_step' not in conn:
            name = conn['name']
            bindings[conn['input_step']][conn['input']] = task_inputs[name].copy()
            bindings[conn['input_step']][conn['input']]['data'] = inputs[name]['script_data']

    # Traverse analyses in topological order
    for step_set in utils.toposort(dependencies):
        for step in step_set:
            # Visualizations cannot be executed
            if 'visualization' in steps[step] and steps[step]['visualization']:
                continue

            # Run step
            print '--- beginning: %s ---' % steps[step]['name']
            out = core_run(steps[step]['task'], bindings[step])
            print '--- finished: %s ---' % steps[step]['name']
            # Update bindings of downstream analyses
            if step in downstream:
                for name, conn_list in downstream[step].iteritems():
                    for conn in conn_list:
                        if 'input_step' in conn:
                            # This is a connection to a downstream step
                            b = bindings[conn['input_step']]
                            b[conn['input']] = out[name]
                        else:
                            # This is a connection to a final output
                            o = outputs[conn['name']]
                            o['script_data'] = out[name]['data']

    # Output visualization parameters
    outputs['_visualizations'] = []
    for step in task['steps']:
        if 'visualization' not in step or not step['visualization']:
            continue
        vis_bindings = {}
        for b, value in bindings[step['name']].iteritems():
            value['script_data'] = value['data']
            vis_input = None
            for step_input in step['task']['inputs']:
                if step_input['name'] == b:
                    vis_input = step_input

            if not vis_input:
                raise Exception('Could not find visualization input named %s.' % b)

            info = {
                'task': task,
                'task_inputs': task_inputs,
                'task_outputs': task_outputs,
                'mode': value.get('mode'),
                'inputs': inputs,
                'outputs': outputs,
                'status': kwargs.get('status'),
                'job_mgr': kwargs.get('_job_manager'),
                'kwargs': kwargs
            }

            e = events.trigger('run.handle_output', {
                'info': info,
                'task_output': vis_input,
                'output': value,
                'outputs': {step['name']: value},
                'name': step['name']
            })

            if not e.default_prevented and 'mode' in value:
                io.push(value['data'], value)
            else:
                vis_bindings[b] = {
                    'type': vis_input['type'],
                    'format': vis_input['format'],
                    'data': value['data']
                }

            """
            # Validate the output
            if (validate and not girder_worker.core.isvalid(
                    vis_input['type'], value)):
                raise Exception(
                    'Output %s (%s) is not in the expected type (%s) and '
                    'format (%s).' % (
                        name, type(value['data']),
                        vis_input['type'], value['format']))

            if auto_convert:
                vis_bindings[b] = girder_worker.core.convert(
                    vis_input['type'],
                    value,
                    {'format': vis_input['format']}
                )

            elif value['format'] == vis_input['format']:
                data = script_output['data']
                if 'mode' in value:
                    girder_worker.core.io.push(data, value)
                else:
                    vis_bindings[b] = {
                        'type': vis_input['type'],
                        'format': vis_input['format'],
                        'data': data
                    }
            else:
                raise Exception(
                    'Expected exact format match but "' +
                    value['format'] +
                    '" != "' + vis_input['format'] + '".'
                )
            """

            vis_bindings[b].pop('script_data', None)

        outputs['_visualizations'].append({
            'mode': 'preset',
            'type': step['name'],
            'inputs': vis_bindings
        })



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################
import celery

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.constants import TokenScope
from girder.api.rest import Resource

from ..celery import getCeleryApp


class Worker(Resource):
    def __init__(self):
        super(Worker, self).__init__()
        self.resourceName = 'worker'
        self.route('GET', ('status',), self.getWorkerStatus)

    @autoDescribeRoute(
        Description('Get worker status and task information.')
        .notes('Return -1 if the broker is inaccessible.')
    )
    @access.user(scope=TokenScope.DATA_READ)
    def getWorkerStatus(self):
        app = getCeleryApp()
        result = {}
        conn = app.connection_for_read()
        try:
            conn.ensure_connection(max_retries=1)
        except celery.exceptions.OperationalError:
            return -1

        status = app.control.inspect()
        result['report'] = status.report()
        result['stats'] = status.stats()
        result['ping'] = status.ping()
        result['active'] = status.active()
        result['reserved'] = status.reserved()
        return result



# The R phylo tree format is a list where the elements
# are not guaranteed to be in any particular order.
# Here we determine which element is which.
element_names_sexp = input.do_slot('names')
element_names = [x for x in element_names_sexp]

# required elements
tipLabelIndex = -1
if 'tip.label' in element_names:
    tipLabelIndex = element_names.index('tip.label')
else:
    print 'Error: tip.label not found in input ape tree'

nNodeIndex = -1
if 'Nnode' in element_names:
    nNodeIndex = element_names.index('Nnode')
else:
    print 'Error: Nnode not found in input ape tree'

edgeIndex = -1
if 'edge' in element_names:
    edgeIndex = element_names.index('edge')
else:
    print 'Error: edge not found in input ape tree'

# optional elements
edgeLengthIndex = -1
if 'edge.length' in element_names:
    edgeLengthIndex = element_names.index('edge.length')

nodeLabelIndex = -1
if 'node.label' in element_names:
    nodeLabelIndex = element_names.index('node.label')

leafCount = len(input[tipLabelIndex])
totalNodes = leafCount + int(input[nNodeIndex][0])

nodes = []
nodeMap = {}


def nodeNameFromIndex(index):
    if index < leafCount + 1:
        # node is a taxon, return the species name
        return input[tipLabelIndex][index - 1]
    elif nodeLabelIndex != -1:
        return input[nodeLabelIndex][index - 1 - leafCount]
    return ''


# loop through the nodes and create a dict for each one
for index in range(1, totalNodes + 1):
    node = {'node_data': {}}
    node['node_data']['node name'] = nodeNameFromIndex(index)
    if index > leafCount:
        # node is not a taxon, so add an empty children array
        node['children'] = []
    nodes.append(node)
    nodeMap[index] = node

# go through the edge table and add fields to the nodes in the collection
edgeCount = len(input[edgeIndex])/2

for edge in range(edgeCount):
    startNodeIndex = int(input[edgeIndex][edge])
    endNodeIndex = int(input[edgeIndex][edgeCount + edge])
    startNode = nodeMap[startNodeIndex]
    endNode = nodeMap[endNodeIndex]
    if edgeLengthIndex != -1:
        # add branch length to end node
        try:
            endNode['edge_data'] = {'weight': input[edgeLengthIndex][edge]}
        except TypeError:
            print 'error on edge or no branchlength:', edge
        except IndexError:
            print 'error on edge or no branchlength:', edge

    # add edge leaving start node and going to endnode
    startNode['children'].append(endNode)

output = nodeMap[leafCount + 1]
output['node_fields'] = ['node name', 'node weight']
output['edge_fields'] = ['weight']


def nodeWeights(node, cur):
    weight = node.get('edge_data', {'weight': 0.0})['weight']
    if isinstance(weight, (int, float)):
        cur += weight
    node['node_data']['node weight'] = cur
    for c in node.get('children', []):
        nodeWeights(c, cur)


nodeWeights(output, 0.0)



import girder_client
import json
import os
from girder_worker import config
from six import StringIO

# Make a sensible limit for metadata outputs
MAX_METADATA_LENGTH = 4 * 1024 * 1024  # 4MB


def _get_cache_settings(spec):
    if not spec.get('use_cache', True):
        return None
    if not config.getboolean('girder_io', 'diskcache_enabled'):
        return None
    return dict(
        directory=config.get('girder_io', 'diskcache_directory'),
        eviction_policy=config.get('girder_io', 'diskcache_eviction_policy'),
        size_limit=config.getint('girder_io', 'diskcache_size_limit'),
        cull_limit=config.getint('girder_io', 'diskcache_cull_limit'),
        large_value_threshold=config.getint(
            'girder_io', 'diskcache_large_value_threshold'),
    )


def _init_client(spec, require_token=False):
    if 'api_url' in spec:
        client = girder_client.GirderClient(
            apiUrl=spec['api_url'],
            cacheSettings=_get_cache_settings(spec))
    elif 'host' in spec:
        scheme = spec.get('scheme', 'http')
        port = spec.get('port', {
            'http': 80,
            'https': 443
        }[scheme])
        api_root = spec.get('api_root', '/api/v1')
        client = girder_client.GirderClient(
            host=spec['host'], scheme=scheme, apiRoot=api_root, port=port,
            cacheSettings=_get_cache_settings(spec))
    else:
        raise Exception('You must pass either an api_url or host key for '
                        'Girder input and output bindings.')

    if 'token' in spec:
        client.setToken(spec['token'])
    elif require_token:
        raise Exception('You must pass a token for Girder authentication.')

    return client


def _fetch_parent_item(file_id, client, dest):
    """
    Fetches the whole item that contains the given file ID into the given
    destination directory. Returns the path to the specific file once the
    download is complete.
    """
    target_file = client.getResource('file', file_id)
    item = client.getResource('item', target_file['itemId'])
    dest = os.path.join(dest, client.transformFilename(item['name']))
    target_path = dest

    for file in client.listFile(item['_id']):
        path = os.path.join(dest, client.transformFilename(file['name']))
        client.downloadFile(file['_id'], path)

        if file['_id'] == target_file['_id']:
            target_path = path

    return target_path


def fetch_handler(spec, **kwargs):
    resource_type = spec.get('resource_type', 'file').lower()
    task_input = kwargs.get('task_input', {})
    target = task_input.get('target', 'filepath')
    fetch_parent = spec.get('fetch_parent', False)
    direct_path = spec.get('direct_path')

    if 'id' not in spec:
        raise Exception('Must pass a resource ID for girder inputs.')
    if 'name' not in spec:
        raise Exception('Must pass a name for girder inputs.')

    client = _init_client(spec)
    filename = client.transformFilename(spec['name'])
    dest = os.path.join(kwargs['_tempdir'], filename)

    if resource_type == 'folder':
        client.downloadFolderRecursive(spec['id'], dest)
    elif resource_type == 'item':
        client.downloadItem(spec['id'], kwargs['_tempdir'], filename)
    elif resource_type == 'file':
        if fetch_parent:
            # If we fetch the parent, we can't use direct paths as the
            # task may needs all of the siblings next to each other
            dest = _fetch_parent_item(spec['id'], client, kwargs['_tempdir'])
        elif (direct_path and config.getboolean('girder_io', 'allow_direct_path') and
                os.path.isfile(direct_path)):
            # If the specification includes a direct path AND it is allowed by
            # the worker configuration AND it is a reachable file, use it.
            dest = direct_path
        else:
            client.downloadFile(spec['id'], dest)
    else:
        raise Exception('Invalid resource type: ' + resource_type)

    if target == 'filepath':
        return dest
    elif target == 'memory':
        with open(dest, 'rb') as fd:
            return fd.read()
    else:
        raise Exception('Invalid Girder push target: ' + target)


def _require_name(name, spec):
    if not name and not spec.get('name'):
        raise Exception('Girder output missing "name" field.')
    return name or spec['name']


def _send_to_girder(client, spec, stream, size, reference, name=None):
    """
    Send an output to Girder as either a file or as metadata on an item
    or folder.

    :param client: The Girder client.
    :param spec: The output binding.
    :param stream: The stream holding the contents of the output data.
    :param size: Size in bytes of the output data
    :param reference: Arbitrary reference string to be passed to the server.
    :param name: Name of the created resource (not required for metadata outputs).
    """
    if spec.get('as_metadata') is True:
        if size > MAX_METADATA_LENGTH:
            raise Exception('Girder metadata output too large (%s bytes).' % size)

        try:
            obj = json.load(stream)
        except (ValueError, TypeError):
            raise Exception('Invalid JSON for metadata output')

        if not isinstance(obj, dict):
            raise Exception('Output must be a JSON Object (got type %s).' % type(obj))

        if 'parent_id' in spec:
            # We need to create a new item and attach this metadata to it
            name = _require_name(name, spec)
            itemId = client.createItem(parentFolderId=spec['parent_id'], name=name)['_id']
            client.addMetadataToItem(itemId, obj)
        elif 'item_id' in spec:
            # We need to attach this as metadata on an existing resource
            client.addMetadataToItem(spec['item_id'], obj)
        else:
            raise Exception('Girder metadata outputs require a parent_id or item_id.')
    else:
        if 'parent_id' not in spec:
            raise Exception('Must pass parent ID for Girder file outputs.')

        name = _require_name(name, spec)

        parent_type = spec.get('parent_type', 'folder')
        file = client.uploadFile(
            parentId=spec['parent_id'], stream=stream, size=size, parentType=parent_type,
            name=name, reference=reference)

        if 'metadata' in spec:
            client.addMetadataToItem(file['itemId'], spec['metadata'])


def push_handler(data, spec, **kwargs):
    reference = spec.get('reference')

    if reference is None:
        # Check for reference in the job manager if none in the output spec
        reference = getattr(kwargs.get('_job_manager'), 'reference', None)

    task_output = kwargs.get('task_output', {})
    target = task_output.get('target', 'filepath')

    client = _init_client(spec, require_token=True)

    if target == 'memory':
        _send_to_girder(
            client=client, spec=spec, stream=StringIO(data), size=len(data), reference=reference)
    elif target == 'filepath':
        name = spec.get('name') or os.path.basename(data)
        if os.path.isdir(data):
            if spec['parent_type'] == 'item':
                for f in os.listdir(data):
                    path = os.path.join(data, f)
                    if os.path.isfile(path):
                        client.uploadFileToItem(spec['parent_id'], path, reference=reference)
            else:
                client.upload(data, spec['parent_id'], spec['parent_type'], reference=reference)
        else:
            size = os.path.getsize(data)
            with open(data, 'rb') as fd:
                _send_to_girder(
                    client=client, spec=spec, stream=fd, size=size, reference=reference, name=name)
    else:
        raise Exception('Invalid Girder push target: ' + target)


def load(params):
    from girder_worker.core import io
    io.register_fetch_handler('girder', fetch_handler)
    io.register_push_handler('girder', push_handler)



import os
import zipfile


output = os.path.splitext(input)[0]

try:
    os.makedirs(output)
except OSError:
    if not os.path.exists(output):
        raise

with zipfile.ZipFile(input) as zf:
    zf.extractall(output)



"""
This module contains the worker events framework. Listeners should subscribe
to events by calling, e.g.

    ``girder_worker.events.bind('event.name', 'my.handler', handler_function)``

And events should be published like so:

    ``e = girder_worker.events.trigger('event.name', info)``

The variable ``e`` would be the event that was created, which includes any
responses that were added by subscribers to the event.
"""

_mapping = {}


class Event(object):
    """
    An Event object is created when an event is triggered. It is passed to
    each of the listeners of the event, which have a chance to add information
    to the event, and also optionally stop the event from being further
    propagated to other listeners, and also optionally instruct the caller that
    it should not execute its default behavior.
    """

    def __init__(self, name, info):
        self.name = name
        self.info = info
        self.propagate = True
        self.default_prevented = False
        self.responses = []

    def prevent_default(self):
        self.default_prevented = True
        return self

    def stop_propagation(self):
        self.propagate = False
        return self

    def add_response(self, response):
        self.responses.append(response)
        return self


def bind(event_name, handler_name, handler):
    global _mapping
    if event_name not in _mapping:
        _mapping[event_name] = []

    _mapping[event_name].append({
        'name': handler_name,
        'handler': handler
    })


def unbind(event_name, handler_name):
    global _mapping
    if event_name not in _mapping:
        return

    for handler in _mapping[event_name]:
        if handler['name'] == handler_name:
            _mapping[event_name].remove(handler)
            break


def trigger(event_name, info=None):
    global _mapping
    e = Event(event_name, info)
    for handler in _mapping.get(event_name, ()):
        handler['handler'](e)

        if e.propagate is False:
            break

    return e



import bson

#  *********************************************
#  **** converting from APE to Arbor below *****
#  *********************************************


def printApeTree(apeTree):
    print 'print overall apeTree:'
    print apeTree
    print '[0]:', apeTree[0]
    print '[1]:', apeTree[1]
    print '[2]:', apeTree[2]
    print '[3]:', apeTree[3]


# lookup taxa by name so we don't get the nodes scrambled in the APE tree
def returnNodeNameFromIndex(apeTree, index):
    if (len(apeTree[1]) == 1):
        leafIndex = 2
    else:
        # print 'alternative ape tree compondent order case'
        leafIndex = 1
    leafCount = len(apeTree[leafIndex])
    # print 'lookup for index:',index
    if index < leafCount+1:
        # node is a taxon, return the species name
        nodeName = apeTree[leafIndex][index-1]
    else:
        nodeName = 'node'+str(index)
    return nodeName

    # sometimes the tuples in phylo class instance are re-orderered.
    # the transformed tree seems to have the order:
    # [0]=edges, [1]=tiplabels, [2]=internalNodeCount, [3]= branchlengths,
    # which has [1] and [2] switched
    # from the definition.  We will look and switch them if necessary


def addApeTreeNodesIntoTreeStore(apeTree, data_coll):
    if (len(apeTree[1]) == 1):
        leafIndex = 2
        countIndex = 1
    else:
        # print 'alternative ape tree compondent order case'
        leafIndex = 1
        countIndex = 2
    # calculate how many nodes are here
    leafCount = len(apeTree[leafIndex])
    # printApeTree(apeTree)
    # print 'apeTree[countIndex]:',apeTree[countIndex]
    # print 'apeTree[countIndex][0]:',apeTree[countIndex][0]
    totalNodes = leafCount + int(apeTree[countIndex][0])
    # loop through the nodes and create a document for each one
    for index in range(1, totalNodes+1):
        node = dict()
        node['name'] = returnNodeNameFromIndex(apeTree, index)
        if index > leafCount:
            # node is not a taxon, so add an empty clade array
            node['clades'] = []
        data_coll.insert(node)


# traverse the edges of an apeTree and add the connectivity from the apeTree
# into the mongo collection representing the tree.  This method assumes the
# nodes have  been added previously.
def addApeTreeEdgesIntoTreeStore(apeTree, data_coll):
    # go through the edge table and add fields to the nodes in the collection
    edgeCount = len(apeTree[0])/2
    # print 'found edge count to be: ',edgeCount
    # print 'edges:',apeTree[3]
    for edgeIndex in range(0, edgeCount):
        startNodeIndex = int(apeTree[0][edgeIndex])
        endNodeIndex = int(apeTree[0][edgeCount+edgeIndex])
        startNodeQuery = {
            'name': returnNodeNameFromIndex(apeTree, startNodeIndex)}
        endNodeQuery = {'name': returnNodeNameFromIndex(apeTree, endNodeIndex)}
        startNode = data_coll.find_one(startNodeQuery)
        endNode = data_coll.find_one(endNodeQuery)
        # print 'edgeIndex: ',edgeIndex,'endnode:  ', endNode
        # add branch length to end node
        try:
            endNode['branch_length'] = apeTree[3][edgeIndex]
        except TypeError:
            print 'error on edgeIndex or no branchlength:', edgeIndex

        # print 'edgeIndex: ',edgeIndex,'endnode:  ', endNode
        data_coll.update(endNodeQuery, endNode)
        # add edge leaving start node and going to endnode
        startNode['clades'].append(endNode['_id'])
        data_coll.update(startNodeQuery, startNode)


# The Arbor treestore already has all its tree nodes, find the root of the tree
# and add a handle node into the collection pointing to the root node.
def addHandleNodeIntoTreeStore(apeTree, data_coll):
    if (len(apeTree[1]) == 1):
        leafIndex = 2
    else:
        # print 'alternative ape tree compondent order case'
        leafIndex = 1
    leafCount = len(apeTree[leafIndex])
    # add the 'handle node' to the collection.  Find the root of the dataset by
    # getting the node immediately after the last leaf
    rootNodeQuery = {'name': returnNodeNameFromIndex(apeTree, leafCount+1)}
    rootNode = data_coll.find_one(rootNodeQuery)
    handlenode = dict()
    # handlenode['name'] = ''
    handlenode['rooted'] = True
    handlenode['clades'] = []
    handlenode['clades'].append(rootNode['_id'])
    data_coll.insert(handlenode)


# this erases the name attributes on internal nodes of the tree.  Names for
# taxon nodes are not affected
def clearInternalNodeNames(apeTree, data_coll):
    if (len(apeTree[1]) == 1):
        leafIndex = 2
        countIndex = 1
    else:
        print 'alternative ape tree compondent order case'
        leafIndex = 1
        countIndex = 2
    leafCount = len(apeTree[leafIndex])
    totalNodes = leafCount + int(apeTree[countIndex][0])
    for index in range(1, totalNodes+1):
        if index > leafCount:
            # node is not a taxon, so remove its name entry
            nodeQuery = {'name': returnNodeNameFromIndex(apeTree, index)}
            internalNode = data_coll.find_one(nodeQuery)
            if 'name' in internalNode:
                del internalNode['name']
            data_coll.update(nodeQuery, internalNode)


# perform the steps that transform an apeTree (passed through the parameter
# variable) into a document collection stored in the Arbor TreeStore.
# The steps are: (1)Nodes are created, (2) edges added, (3) handled node is
# created in the collection, and (4) names are erased for hierarchical nodes
def importApeTreeToArbor(apeTree, data_coll):
    data_coll.drop()
    global global_next_node
    global global_next_edge
    global global_leafname_list
    global_next_node = 0
    global_next_edge = 0
    global_leafname_list = []
    addApeTreeNodesIntoTreeStore(apeTree, data_coll)
    addApeTreeEdgesIntoTreeStore(apeTree, data_coll)
    addHandleNodeIntoTreeStore(apeTree, data_coll)
    clearInternalNodeNames(apeTree, data_coll)


# Adapting treestore conversion to girder_worker
class InMemoryCollection:
    items = []
    nameMap = {}
    currentId = 0

    def drop(self):
        self.items = []
        self.nameMap = {}
        self.currentId = 0

    def insert(self, item):
        item['_id'] = self.currentId
        self.currentId += 1
        self.items.append(item)
        if 'name' in item:
            self.nameMap[item['name']] = item

    def update(self, query, item):
        pass

    def find_one(self, query):
        return self.nameMap[query['name']]


c = InMemoryCollection()
importApeTreeToArbor(input, c)

output = ''.join([bson.BSON.encode(d) for d in c.items])



"""Define the common parent type for all json specs contained in the package."""

import six
import json
from collections import Mapping
import copy


class SpecMixin(object):
    """An abstract mixin class implementing the Spec methods."""

    __defaults = {}

    def __init__(self, *args, **kw):
        """Initialize a new spec object.

        When called with no arguments, constructs an empty mapping.
        >>> Spec()
        {}

        Positional arguments can be mappings
        >>> Spec({})
        {}

        or JSON formatted strings.
        >>> Spec('{}')
        {}

        Multiple positional arguments are applied in order.
        >>> Spec({'a': 0}, '{"b": 1}', {'a': 2})
        {"a": 2, "b": 1}

        Items given as keyword argument will be insert last.
        >>> Spec({'a': 0}, '{"b": 1}', a=2, c=3)
        {"a": 2, "b": 1, "c": 3}
        """
        # add internal validation store
        self.__checks = {}

        self.add_validation_check('Spec.json', Spec.__ensure_json)

        # process positional arguments
        for arg in args:
            if isinstance(arg, six.string_types):
                arg = json.loads(arg)
            d = dict(arg)
            self.update(d)

        # process keyword arguments
        self.update(kw)

    def __copy__(self):
        return copy.copy(dict(self))

    def __deepcopy__(self, memodict={}):
        return copy.deepcopy(dict(self))

    @staticmethod
    def _serializer(value):
        """Serialize the value or raise a TypeError."""
        return value

    def __check(self, key=None, oldvalue=None, newvalue=None, **kw):
        """Call all registered validation methods on this spec."""
        for func in six.itervalues(self.__checks):
            func(self, key, oldvalue, newvalue)

    def __ensure_json(self, key=None, oldvalue=None, newvalue=None, **kw):
        """Raise a ValueError if the spec is not valid JSON."""
        if key is None:
            for key, value in six.iteritems(self):
                self.__ensure_json(key,
                                   newvalue=value,
                                   default=self._serializer)
        elif not isinstance(key, six.string_types):
            raise TypeError('Spec keys must be string typed.')

        if newvalue is not None:
            try:
                json.dumps(newvalue, default=self._serializer)
            except Exception:
                raise ValueError(
                    '"%s" is not valid json.' % repr(newvalue)
                )

    def update(self, other=None, **kw):
        """A recursive version of ``dict.update``."""
        if other is not None:
            _update(self, dict(other))
        _update(self, kw)

    def check(self, *args, **kw):
        """Public validation of the specification.

        This works like the object mutation validation, but the call
        to the handers is made with no key.
        """
        self.__check(*args, **kw)

    def json(self, **kw):
        """Return the spec as a json blob.

        Keyword arguments are passed to :py:func:`json.dumps`.
        """
        kw.setdefault('default', self._serializer)
        self.check()
        return json.dumps(self, **kw)

    def __str__(self):
        """Serialize the object as json."""
        return self.json(default=str, skipkeys=True, sort_keys=True)

    def __repr__(self):
        """Serialize the object as json using repr."""
        return self.json(default=repr, skipkeys=True, sort_keys=True)

    def add_validation_check(self, name, func):
        """Add a method used to validate spec objects on mutation.

        :param str name: A name given to the check
        :param function func: The function to call

        Validation functions should have the following signature:

        func(self, key=None, oldvalue=None, newvalue=None, **kw)

        * self: the spec object itself
        * key: the key that changed
        * oldvalue: the old value of the key
        * newvalue: the new value of the key

        When a key is given the function can assume that all other keys in the
        spec are valid.  Old and new values of the key are given when
        applicable.  The validation function will be called whenever the spec
        is changed or on demand via the public interface :py:method:`check`.
        """
        self.__checks[name] = func

    def remove_validation_check(self, name):
        """Remove a previously added validation method."""
        self.__checks.pop(name)

    @classmethod
    def make_property(cls, name,
                      doc=None, default=None):
        """Expose a spec item as a class attribute.

        :param str name: The property name
        :param str doc: The docstring to add
        :param object default: The default value of the property
        :param function init: A constructor called on all new values

        A sample property definition
        >>> class MySpec(Spec): pass
        >>> MySpec.make_property('name', 'This is the name', 'Bob')

        Properties can be set via the Spec constructor, as an item,
        or as an attribute.  Standard attribute methods are supported.
        >>> s = MySpec(name="Roger")
        >>> s.name = 'Alice'
        >>> s.name == s['name']
        True
        >>> del s['name']
        >>> s.name
        'Bob'
        >>> s.name == s['name']
        True

        Default values are returned for unset properties
        >>> 'name' in s
        False
        >>> s['name']
        'Bob'

        Properties can be used as normal attributess

        Optionally properties can have docstrings
        >>> MySpec.name.__doc__
        'This is the name'
        """
        prop = property(
            lambda self: self.__getitem__(name),
            lambda self, value: self.__setitem__(name, value),
            lambda self: self.__delitem__(name),
            doc
        )
        cls.__defaults[name] = default
        setattr(cls, name, prop)

    def __missing__(self, key):
        """Return default for missing keys."""
        return self.__defaults.get(key, None)


class Spec(SpecMixin, dict):
    r"""
    Defines core utility methods that all spec objects have in common.

    Supports dict-like initialization.

    >>> a = Spec({'a': 1, 'b': {'c': [1, 2, None]}})
    >>> b = Spec(a=1, b={'c': [1, 2, None]})
    >>> a == b
    True

    Also supports initialization from json.

    >>> c = Spec('{"a": 1, "b": {"c": [1, 2, null]}}')
    >>> a == c
    True

    Multiple initialization method can be used together, which will be inserted
    in order.

    >>> Spec('{"a": 0}', {'a': 1}, a=2)
    {"a": 2}

    Updating merging specs is always done recursively.

    >>> Spec('{"a": {"b": 0}}', a={'c': 1})
    {"a": {"b": 0, "c": 1}}

    Conflicts are resolved by taking the value with highest priority (i.e.
    the once provided next in the constructor.)

    >>> Spec('{"a": {"b": 0}}', {"a": []})
    {"a": []}
    >>> Spec('{"a": {"b": 0}}', {"a": []}, a={"c": 1})
    {"a": {"c": 1}}
    >>> Spec('{"a": []}', {"a": {"b": 0}}, a={"c": 1})
    {"a": {"b": 0, "c": 1}}

    Serialization is performed as json

    >>> str(a)
    '{"a": 1, "b": {"c": [1, 2, null]}}'

    Strings are assumed to be utf-8 encoded.

    >>> str(Spec({u"for\u00eat": u"\ud83c\udf33 \ud83c\udf32 \ud83c\udf34"}))
    '{"for\\u00eat": "\\ud83c\\udf33 \\ud83c\\udf32 \\ud83c\\udf34"}'

    Methods that mutate the state of the Spec will test if the new state is
    valid, restoring the original state before raising an exception.

    >>> s = Spec({'a': 0})
    >>> try:
    ...     s['a'] = object
    ... except Exception:
    ...     pass
    ... else:
    ...     assert False
    >>> s
    {"a": 0}

    Spec constructors are idempotent

    >>> Spec(a='a') == Spec(Spec(a='a'))
    True
    """

    def __setitem__(self, key, value):
        """Call ``dict.__setitem__`` and restore if the result is invalid."""
        restore = False
        old = None
        if key in self:
            restore = True
            old = super(Spec, self).__getitem__(key)
        try:
            super(Spec, self).__setitem__(key, value)
            self.check(key=key, oldvalue=old, newvalue=value)
        except Exception:
            if restore:
                super(Spec, self).__setitem__(key, old)
            else:
                super(Spec, self).__delitem__(key)
            raise

    def __delitem__(self, key):
        """Call ``dict.__delitem__`` and restore if the result is invalid."""
        restore = False
        if key in self:
            restore = True
            old = super(Spec, self).__getitem__(key)
        try:
            super(Spec, self).__delitem__(key)
            self.check(key=key, oldvalue=old)
        except Exception:
            if restore:
                super(Spec, self).__setitem__(key, old)
            raise


def _update(d, u):
    """Update a nested dict recursively in place.

    >>> _update({'a': 0, 'b': 2}, {'a': 1})
    {'a': 1, 'b': 2}

    >>> _update({'a': {'b': 1}}, {'a': 0})
    {'a': 0}

    >>> d = _update({'a': {'b': 1}}, {'a': {'c': 2}})
    >>> d == {'a': {'b': 1, 'c': 2}}
    True
    """
    for k, v in six.iteritems(u):
        r = d.get(k, {})
        if isinstance(v, Mapping) and isinstance(r, Mapping):
            _update(r, v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


__all__ = ('Spec',)



import rpy2.robjects


def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    env = rpy2.robjects.globalenv

    # Clear out workspace variables and packages
    rpy2.robjects.reval("""
        rm(list = ls())
        pkgs <- names(sessionInfo()$otherPkgs)
        if (!is.null(pkgs)) {
            pkgs <- paste('package:', pkgs, sep = '')
            lapply(pkgs, detach, character.only = TRUE, unload = TRUE)
        }
        """, env)

    env['tempdir'] = kwargs.get('_tempdir')

    for name in inputs:
        env[str(name)] = inputs[name]['script_data']

    rpy2.robjects.reval(task['script'], env)

    for name, task_output in task_outputs.iteritems():
        d = outputs[name]
        d['script_data'] = env[str(name)]

        # Hack to detect scalar values from R.
        # The R value might not have a len() so wrap in a try/except.
        try:
            if len(d['script_data']) == 1:
                d['script_data'] = d['script_data'][0]
        except TypeError:
            pass
